@@ -1,6 +1,6 @@
 {- adjusted branch
  -
- - Copyright 2016-2020 Joey Hess <id@joeyh.name>
+ - Copyright 2016-2023 Joey Hess <id@joeyh.name>
  -
  - Licensed under the GNU AGPL version 3 or higher.
  -}
@@ -67,6 +67,7 @@ import Annex.CurrentBranch
 import Types.CleanupActions
 import qualified Database.Keys
 import Config
+import Logs.View (is_branchView)
 
 import qualified Data.Map as M
 
@@ -207,7 +208,7 @@ enterAdjustedBranch adj = inRepo Git.Branch.current >>= \case
 	go currbranch = do
 		let origbranch = fromAdjustedBranch currbranch
 		let adjbranch = adjBranch $ originalToAdjusted origbranch adj
-		ifM (inRepo (Git.Ref.exists adjbranch) <&&> (not <$> Annex.getState Annex.force))
+		ifM (inRepo (Git.Ref.exists adjbranch) <&&> (not <$> Annex.getRead Annex.force) <&&> pure (not (is_branchView origbranch)))
 			( do
 				mapM_ (warning . unwords)
 					[ [ "adjusted branch"
@@ -459,10 +460,10 @@ findAdjustingCommit (AdjBranch b) = go =<< catCommit b
 			_ -> return Nothing
 
 {- Check for any commits present on the adjusted branch that have not yet
- - been propigated to the basis branch, and propigate them to the basis
+ - been propigated to the basis branch, and propagate them to the basis
  - branch and from there on to the orig branch.
  -
- - After propigating the commits back to the basis banch,
+ - After propigating the commits back to the basis branch,
  - rebase the adjusted branch on top of the updated basis branch.
  -}
 propigateAdjustedCommits :: OrigBranch -> Adjustment -> Annex ()
@@ -490,7 +491,9 @@ propigateAdjustedCommits' origbranch adj _commitsprevented =
 						, rebase currcommit newparent
 						)
 			Nothing -> return (Nothing, return ())
-		Nothing -> return (Nothing, return ())
+		Nothing -> do
+			warning $ "Cannot find basis ref " ++ fromRef basis ++ "; not propagating adjusted commits to original branch " ++ fromRef origbranch
+			return (Nothing, return ())
   where
 	(BasisBranch basis) = basisBranch adjbranch
 	adjbranch@(AdjBranch currbranch) = originalToAdjusted origbranch adj
@@ -533,7 +536,7 @@ rebaseOnTopMsg = "rebasing adjusted branch on top of updated original branch"
 reverseAdjustedCommit :: Sha -> Adjustment -> (Sha, Commit) -> OrigBranch -> Annex (Either String Sha)
 reverseAdjustedCommit commitparent adj (csha, basiscommit) origbranch
 	| length (commitParent basiscommit) > 1 = return $
-		Left $ "unable to propigate merge commit " ++ show csha ++ " back to " ++ show origbranch
+		Left $ "unable to propagate merge commit " ++ show csha ++ " back to " ++ show origbranch
 	| otherwise = do
 		cmode <- annexCommitMode <$> Annex.getGitConfig
 		treesha <- reverseAdjustedTree commitparent adj csha
