@@ -104,6 +104,7 @@ import Utility.Metered
 import qualified Utility.RawFilePath as R
 
 import qualified System.FilePath.ByteString as P
+import System.PosixCompat.Files (isSymbolicLink, linkCount)
 
 {- Prevents the content from being removed while the action is running.
  - Uses a shared lock.
@@ -139,7 +140,8 @@ lockContentShared key a = lockContentUsing lock key notpresent $
  -
  - If locking fails, throws an exception rather than running the action.
  -
- - If locking fails because the the content is not present, runs the
+ - When the content file itself is used as the lock file, 
+ - and locking fails because the the content is not present, runs the
  - fallback action instead. However, the content is not guaranteed to be
  - present when this succeeds.
  -}
@@ -390,9 +392,9 @@ withTmp key action = do
  - with colliding files it's their own fault and B) adding such a check
  - would not catch all cases of colliding keys. For example, perhaps 
  - a remote has a key; if it's then added again with different content then
- - the overall system now has two different peices of content for that
+ - the overall system now has two different pieces of content for that
  - key, and one of them will probably get deleted later. So, adding the
- - check here would only raise expectations that git-annex cannot truely
+ - check here would only raise expectations that git-annex cannot truly
  - meet.
  -
  - May return false, when a particular variety of key is not being
@@ -553,7 +555,7 @@ sendAnnex key rollback sendobject = go =<< prepSendAnnex' key
 {- Returns a file that contains an object's content,
  - and a check to run after the transfer is complete.
  -
- - When a file is unlocked, it's possble for its content to
+ - When a file is unlocked, it's possible for its content to
  - change as it's being sent. The check detects this case
  - and returns False.
  -
@@ -718,7 +720,7 @@ listKeys' keyloc want = do
 saveState :: Bool -> Annex ()
 saveState nocommit = doSideAction $ do
 	Annex.Queue.flush
-	Database.Keys.closeDb
+	Database.Keys.flushDb
 	unless nocommit $
 		whenM (annexAlwaysCommit <$> Annex.getGitConfig) $
 			Annex.Branch.commit =<< Annex.Branch.commitMessage
