@@ -184,10 +184,13 @@ recordImportTree remote importtreeconfig importable = do
 				unlessM (stillpresent db oldkey) $
 					logChange oldkey (Remote.uuid remote) InfoMissing
 			_ -> noop
-		db <- Export.openDb (Remote.uuid remote)
-		forM_ (exportedTreeishes oldexport) $ \oldtree ->
-			Export.runExportDiffUpdater updater db oldtree finaltree
-		Export.closeDb db
+		-- When the remote is versioned, it still contains keys
+		-- that are not present in the new tree.
+		unless (Remote.versionedExport (Remote.exportActions remote)) $ do
+			db <- Export.openDb (Remote.uuid remote)
+			forM_ (exportedTreeishes oldexport) $ \oldtree ->
+				Export.runExportDiffUpdater updater db oldtree finaltree
+			Export.closeDb db
 
 buildImportCommit' :: Remote -> ImportCommitConfig -> Maybe Sha -> History Sha -> Annex (Maybe Sha)
 buildImportCommit' remote importcommitconfig mtrackingcommit imported@(History ti _) =
@@ -412,8 +415,10 @@ importKeys remote importtreeconfig importcontent thirdpartypopulated importablec
 		| importcontent = a
 		| otherwise = reuseVectorClockWhile a
 
-	withciddb = withExclusiveLock gitAnnexContentIdentifierLock .
-		bracket CIDDb.openDb CIDDb.closeDb
+	withciddb a = do
+		cidlck <- calcRepo' gitAnnexContentIdentifierLock
+		withExclusiveLock cidlck $
+			bracket CIDDb.openDb CIDDb.closeDb a
 
 	run cidmap importing db = do
 		largematcher <- largeFilesMatcher
@@ -487,7 +492,7 @@ importKeys remote importtreeconfig importcontent thirdpartypopulated importablec
 	startimport cidmap importing db i@(loc, (cid, _sz)) oldversion largematcher = getcidkey cidmap db cid >>= \case
 		(k:ks) ->
 			-- If the same content was imported before
-			-- yeilding multiple different keys, it's not clear
+			-- yielding multiple different keys, it's not clear
 			-- which is best to use this time, so pick the
 			-- first in the list. But, if any of them is a
 			-- git sha, use it, because the content must
@@ -598,7 +603,7 @@ importKeys remote importtreeconfig importcontent thirdpartypopulated importablec
 			let af = AssociatedFile (Just f)
 			let downloader p' tmpfile = do
 				_ <- Remote.retrieveExportWithContentIdentifier
-					ia loc cid (fromRawFilePath tmpfile)
+					ia loc [cid] (fromRawFilePath tmpfile)
 					(Left k)
 					(combineMeterUpdate p' p)
 				ok <- moveAnnex k af tmpfile
@@ -616,7 +621,7 @@ importKeys remote importtreeconfig importcontent thirdpartypopulated importablec
 	doimportsmall cidmap db loc cid sz p = do
 		let downloader tmpfile = do
 			(k, _) <- Remote.retrieveExportWithContentIdentifier
-				ia loc cid (fromRawFilePath tmpfile)
+				ia loc [cid] (fromRawFilePath tmpfile)
 				(Right (mkkey tmpfile))
 				p
 			case keyGitSha k of
@@ -639,7 +644,7 @@ importKeys remote importtreeconfig importcontent thirdpartypopulated importablec
 		let af = AssociatedFile (Just f)
 		let downloader tmpfile p = do
 			(k, _) <- Remote.retrieveExportWithContentIdentifier
-				ia loc cid (fromRawFilePath tmpfile)
+				ia loc [cid] (fromRawFilePath tmpfile)
 				(Right (mkkey tmpfile))
 				p
 			case keyGitSha k of
