@@ -1,6 +1,6 @@
 {- git-annex assistant commit thread
  -
- - Copyright 2012, 2019 Joey Hess <id@joeyh.name>
+ - Copyright 2012-2021 Joey Hess <id@joeyh.name>
  -
  - Licensed under the GNU AGPL version 3 or higher.
  -}
@@ -20,11 +20,11 @@ import Assistant.Drop
 import Types.Transfer
 import Logs.Location
 import qualified Annex.Queue
-import qualified Git.LsFiles
 import Utility.ThreadScheduler
 import qualified Utility.Lsof as Lsof
 import qualified Utility.DirWatcher as DirWatcher
 import Types.KeySource
+import Types.Command
 import Config
 import Annex.Content
 import Annex.Ingest
@@ -32,39 +32,40 @@ import Annex.Link
 import Annex.Perms
 import Annex.CatFile
 import Annex.InodeSentinal
-import Annex.Version
 import Annex.CurrentBranch
+import Annex.FileMatcher
 import qualified Annex
 import Utility.InodeCache
-import Annex.Content.Direct
 import qualified Database.Keys
 import qualified Command.Sync
-import qualified Git.Branch
 import Utility.Tuple
 import Utility.Metered
+import qualified Utility.RawFilePath as R
 
 import Data.Time.Clock
 import qualified Data.Set as S
 import qualified Data.Map as M
 import Data.Either
 import Control.Concurrent
+import System.PosixCompat.Files (fileID, deviceID, fileMode)
 
 {- This thread makes git commits at appropriate times. -}
 commitThread :: NamedThread
 commitThread = namedThread "Committer" $ do
-	havelsof <- liftIO $ inPath "lsof"
+	havelsof <- liftIO $ inSearchPath "lsof"
 	delayadd <- liftAnnex $
-		maybe delayaddDefault (return . Just . Seconds)
-			=<< annexDelayAdd <$> Annex.getGitConfig
+		fmap Seconds . annexDelayAdd <$> Annex.getGitConfig
+	largefilematcher <- liftAnnex largeFilesMatcher
 	msg <- liftAnnex Command.Sync.commitMsg
 	lockdowndir <- liftAnnex $ fromRepo gitAnnexTmpWatcherDir
 	liftAnnex $ do
 		-- Clean up anything left behind by a previous process
 		-- on unclean shutdown.
-		void $ liftIO $ tryIO $ removeDirectoryRecursive lockdowndir
+		void $ liftIO $ tryIO $ removeDirectoryRecursive
+			(fromRawFilePath lockdowndir)
 		void $ createAnnexDirectory lockdowndir
 	waitChangeTime $ \(changes, time) -> do
-		readychanges <- handleAdds lockdowndir havelsof delayadd $
+		readychanges <- handleAdds (fromRawFilePath lockdowndir) havelsof largefilematcher delayadd $
 			simplifyChanges changes
 		if shouldCommit False time (length readychanges) readychanges
 			then do
@@ -94,7 +95,7 @@ refill cs = do
  - runs an action to commit them. If more changes arrive while this is
  - going on, they're handled intelligently, batching up changes into
  - large commits where possible, doing rename detection, and
- - commiting immediately otherwise. -}
+ - committing immediately otherwise. -}
 waitChangeTime :: (([Change], UTCTime) -> Assistant Int) -> Assistant ()
 waitChangeTime a = waitchanges 0
   where
@@ -235,26 +236,14 @@ commitStaged msg = do
 	case v of
 		Left _ -> return False
 		Right _ -> do
-			ok <- Command.Sync.commitStaged Git.Branch.AutomaticCommit msg
+			cmode <- annexCommitMode <$> Annex.getGitConfig
+			ok <- Command.Sync.commitStaged cmode msg
 			when ok $
 				Command.Sync.updateBranches =<< getCurrentBranch
 			return ok
 
-{- OSX needs a short delay after a file is added before locking it down,
- - as pasting a file seems to try to set file permissions or otherwise
- - access the file after closing it. -}
-delayaddDefault :: Annex (Maybe Seconds)
-#ifdef darwin_HOST_OS
-delayaddDefault = ifM (isDirect <||> versionSupportsUnlockedPointers)
-	( return Nothing
-	, return $ Just $ Seconds 1
-	)
-#else
-delayaddDefault = return Nothing
-#endif
-
 {- If there are PendingAddChanges, or InProcessAddChanges, the files
- - have not yet actually been added to the annex, and that has to be done
+ - have not yet actually been added, and that has to be done
  - now, before committing.
  -
  - Deferring the adds to this point causes batches to be bundled together,
@@ -272,133 +261,107 @@ delayaddDefault = return Nothing
  - Any pending adds that are not ready yet are put back into the ChangeChan,
  - where they will be retried later.
  -}
-handleAdds :: FilePath -> Bool -> Maybe Seconds -> [Change] -> Assistant [Change]
-handleAdds lockdowndir havelsof delayadd cs = returnWhen (null incomplete) $ do
+handleAdds :: FilePath -> Bool -> GetFileMatcher -> Maybe Seconds -> [Change] -> Assistant [Change]
+handleAdds lockdowndir havelsof largefilematcher delayadd cs = returnWhen (null incomplete) $ do
 	let (pending, inprocess) = partition isPendingAddChange incomplete
-	direct <- liftAnnex isDirect
-	unlocked <- liftAnnex versionSupportsUnlockedPointers
-	let lockingfiles = not (unlocked || direct)
 	let lockdownconfig = LockDownConfig
-		{ lockingFile = lockingfiles
-		, hardlinkFileTmpDir = Just lockdowndir
+		{ lockingFile = False
+		, hardlinkFileTmpDir = Just (toRawFilePath lockdowndir)
+		, checkWritePerms = True
 		}
-	(pending', cleanup) <- if unlocked || direct
-		then return (pending, noop)
-		else findnew pending
 	(postponed, toadd) <- partitionEithers
-		<$> safeToAdd lockdowndir lockdownconfig havelsof delayadd pending' inprocess
-	cleanup
+		<$> safeToAdd lockdowndir lockdownconfig havelsof delayadd pending inprocess
 
 	unless (null postponed) $
 		refillChanges postponed
 
 	returnWhen (null toadd) $ do
-		added <- addaction toadd $
-			catMaybes <$>
-				if not lockingfiles
-					then addunlocked direct toadd
-					else forM toadd (add lockdownconfig)
-		if DirWatcher.eventsCoalesce || null added || unlocked || direct
-			then return $ added ++ otherchanges
-			else do
-				r <- handleAdds lockdowndir havelsof delayadd =<< getChanges
-				return $ r ++ added ++ otherchanges
+		(toaddannexed, toaddsmall) <- partitionEithers
+			<$> mapM checksmall toadd
+		addsmall toaddsmall
+		addedannexed <- addaction toadd $
+			catMaybes <$> addannexed toaddannexed
+		return $ addedannexed ++ toaddsmall ++ otherchanges
   where
 	(incomplete, otherchanges) = partition (\c -> isPendingAddChange c || isInProcessAddChange c) cs
-	
-	-- Find files that are actually new, and not unlocked annexed
-	-- files. The ls-files is run on a batch of files.
-	findnew [] = return ([], noop)
-	findnew pending@(exemplar:_) = do
-		let segments = segmentXargsUnordered $ map changeFile pending
-		rs <- liftAnnex $ forM segments $ \fs ->
-			inRepo (Git.LsFiles.notInRepo False fs)
-		let (newfiles, cleanup) = foldl'
-			(\(l1, a1) (l2, a2) -> (l1 ++ l2, a1 >> a2))
-			([], return True) rs
-		-- note: timestamp info is lost here
-		let ts = changeTime exemplar
-		return (map (PendingAddChange ts) newfiles, void $ liftIO cleanup)
 
 	returnWhen c a
 		| c = return otherchanges
 		| otherwise = a
 
-	add :: LockDownConfig -> Change -> Assistant (Maybe Change)
-	add lockdownconfig change@(InProcessAddChange { lockedDown = ld }) = 
-		catchDefaultIO Nothing <~> doadd
-	  where
-	  	ks = keySource ld
-		doadd = sanitycheck ks $ do
-			(mkey, mcache) <- liftAnnex $ do
-				showStart "add" $ keyFilename ks
-				ingest nullMeterUpdate (Just $ LockedDown lockdownconfig ks) Nothing
-			maybe (failedingest change) (done change mcache $ keyFilename ks) mkey
-	add _ _ = return Nothing
+	checksmall change =
+		ifM (liftAnnex $ checkFileMatcher largefilematcher (toRawFilePath (changeFile change)))
+			( return (Left change)
+			, return (Right change)
+			)
+
+	addsmall [] = noop
+	addsmall toadd = liftAnnex $ Annex.Queue.addCommand [] "add"
+		[ Param "--force", Param "--"] (map changeFile toadd)
 
 	{- Avoid overhead of re-injesting a renamed unlocked file, by
 	 - examining the other Changes to see if a removed file has the
 	 - same InodeCache as the new file. If so, we can just update
 	 - bookkeeping, and stage the file in git.
 	 -}
-	addunlocked :: Bool -> [Change] -> Assistant [Maybe Change]
-	addunlocked isdirect toadd = do
+	addannexed :: [Change] -> Assistant [Maybe Change]
+	addannexed [] = return []
+	addannexed toadd = do
 		ct <- liftAnnex compareInodeCachesWith
-		m <- liftAnnex $ removedKeysMap isdirect ct cs
+		m <- liftAnnex $ removedKeysMap ct cs
 		delta <- liftAnnex getTSDelta
 		let cfg = LockDownConfig
 			{ lockingFile = False
-			, hardlinkFileTmpDir = Just lockdowndir
+			, hardlinkFileTmpDir = Just (toRawFilePath lockdowndir)
+			, checkWritePerms = True
 			}
 		if M.null m
-			then forM toadd (add cfg)
+			then forM toadd (addannexed' cfg)
 			else forM toadd $ \c -> do
-				mcache <- liftIO $ genInodeCache (changeFile c) delta
+				mcache <- liftIO $ genInodeCache (toRawFilePath (changeFile c)) delta
 				case mcache of
-					Nothing -> add cfg c
+					Nothing -> addannexed' cfg c
 					Just cache ->
 						case M.lookup (inodeCacheToKey ct cache) m of
-							Nothing -> add cfg c
-							Just k -> fastadd isdirect c k
+							Nothing -> addannexed' cfg c
+							Just k -> fastadd c k
 
-	fastadd :: Bool -> Change -> Key -> Assistant (Maybe Change)
-	fastadd isdirect change key = do
+	addannexed' :: LockDownConfig -> Change -> Assistant (Maybe Change)
+	addannexed' lockdownconfig change@(InProcessAddChange { lockedDown = ld }) = 
+		catchDefaultIO Nothing <~> doadd
+	  where
+	  	ks = keySource ld
+		doadd = sanitycheck ks $ do
+			(mkey, _mcache) <- liftAnnex $ do
+				showStart "add" (keyFilename ks) (SeekInput [])
+				ingest nullMeterUpdate (Just $ LockedDown lockdownconfig ks) Nothing
+			maybe (failedingest change) (done change $ fromRawFilePath $ keyFilename ks) mkey
+	addannexed' _ _ = return Nothing
+
+	fastadd :: Change -> Key -> Assistant (Maybe Change)
+	fastadd change key = do
 		let source = keySource $ lockedDown change
-		liftAnnex $ if isdirect
-			then finishIngestDirect key source
-			else finishIngestUnlocked key source
-		done change Nothing (keyFilename source) key
+		liftAnnex $ finishIngestUnlocked key source
+		done change (fromRawFilePath $ keyFilename source) key
 
-	removedKeysMap :: Bool -> InodeComparisonType -> [Change] -> Annex (M.Map InodeCacheKey Key)
-	removedKeysMap isdirect ct l = do
+	removedKeysMap :: InodeComparisonType -> [Change] -> Annex (M.Map InodeCacheKey Key)
+	removedKeysMap ct l = do
 		mks <- forM (filter isRmChange l) $ \c ->
-			catKeyFile $ changeFile c
+			catKeyFile $ toRawFilePath $ changeFile c
 		M.fromList . concat <$> mapM mkpairs (catMaybes mks)
 	  where
 		mkpairs k = map (\c -> (inodeCacheToKey ct c, k)) <$>
-			if isdirect
-				then recordedInodeCache k
-				else Database.Keys.getInodeCaches k
+			Database.Keys.getInodeCaches k
 
 	failedingest change = do
 		refill [retryChange change]
 		liftAnnex showEndFail
 		return Nothing
 
-	done change mcache file key = liftAnnex $ do
+	done change file key = liftAnnex $ do
 		logStatus key InfoPresent
-		ifM versionSupportsUnlockedPointers
-			( do
-				mode <- liftIO $ catchMaybeIO $ fileMode <$> getFileStatus file
-				stagePointerFile file mode =<< hashPointerFile key
-			, do
-				link <- ifM isDirect
-					( calcRepo $ gitAnnexLink file key
-					, makeLink file key mcache
-					)
-				whenM (pure DirWatcher.eventsCoalesce <||> isDirect) $
-					stageSymlink file =<< hashSymlink link
-			)
+		mode <- liftIO $ catchMaybeIO $ fileMode <$> R.getFileStatus (toRawFilePath file)
+		stagePointerFile (toRawFilePath file) mode =<< hashPointerFile key
 		showEndOk
 		return $ Just $ finishedChange change key
 
@@ -406,14 +369,14 @@ handleAdds lockdowndir havelsof delayadd cs = returnWhen (null incomplete) $ do
 	 - and is still a hard link to its contentLocation,
 	 - before ingesting it. -}
 	sanitycheck keysource a = do
-		fs <- liftIO $ getSymbolicLinkStatus $ keyFilename keysource
-		ks <- liftIO $ getSymbolicLinkStatus $ contentLocation keysource
+		fs <- liftIO $ R.getSymbolicLinkStatus $ keyFilename keysource
+		ks <- liftIO $ R.getSymbolicLinkStatus $ contentLocation keysource
 		if deviceID ks == deviceID fs && fileID ks == fileID fs
 			then a
 			else do
 				-- remove the hard link
 				when (contentLocation keysource /= keyFilename keysource) $
-					void $ liftIO $ tryIO $ removeFile $ contentLocation keysource
+					void $ liftIO $ tryIO $ removeFile $ fromRawFilePath $ contentLocation keysource
 				return Nothing
 
 	{- Shown an alert while performing an action to add a file or
@@ -459,7 +422,7 @@ safeToAdd lockdowndir lockdownconfig havelsof delayadd pending inprocess = do
 			else return checked
   where
 	check openfiles change@(InProcessAddChange { lockedDown = ld })
-		| S.member (contentLocation (keySource ld)) openfiles = Left change
+		| S.member (fromRawFilePath (contentLocation (keySource ld))) openfiles = Left change
 	check _ change = Right change
 
 	mkinprocess (c, Just ld) = Just InProcessAddChange
@@ -470,11 +433,11 @@ safeToAdd lockdowndir lockdownconfig havelsof delayadd pending inprocess = do
 
 	canceladd (InProcessAddChange { lockedDown = ld }) = do
 		let ks = keySource ld
-		warning $ keyFilename ks
+		warning $ fromRawFilePath (keyFilename ks)
 			++ " still has writers, not adding"
 		-- remove the hard link
 		when (contentLocation ks /= keyFilename ks) $
-			void $ liftIO $ tryIO $ removeFile $ contentLocation ks
+			void $ liftIO $ tryIO $ removeFile $ fromRawFilePath $ contentLocation ks
 	canceladd _ = noop
 
 	openwrite (_file, mode, _pid)
@@ -493,7 +456,8 @@ safeToAdd lockdowndir lockdownconfig havelsof delayadd pending inprocess = do
 	 -}
 	findopenfiles keysources = ifM crippledFileSystem
 		( liftIO $ do
-			let segments = segmentXargsUnordered $ map keyFilename keysources
+			let segments = segmentXargsUnordered $
+				map (fromRawFilePath . keyFilename) keysources
 			concat <$> forM segments (\fs -> Lsof.query $ "--" : fs)
 		, liftIO $ Lsof.queryDir lockdowndir
 		)
@@ -516,5 +480,5 @@ checkChangeContent change@(Change { changeInfo = i }) =
 			handleDrops "file renamed" present k af []
   where
 	f = changeFile change
-	af = AssociatedFile (Just f)
+	af = AssociatedFile (Just (toRawFilePath f))
 checkChangeContent _ = noop
