@@ -207,13 +207,13 @@ upgradeToDistribution newdir cleanup distributionfile = do
 			unless tarok $
 				error $ "failed to untar " ++ distributionfile
 			sanitycheck $ tmpdir </> installBase
-			installby rename newdir (tmpdir </> installBase)
+			installby R.rename newdir (tmpdir </> installBase)
 		let deleteold = do
 			deleteFromManifest olddir
 			makeorigsymlink olddir
 		return (newdir </> "git-annex", deleteold)
 	installby a dstdir srcdir =
-		mapM_ (\x -> a x (dstdir </> takeFileName x))
+		mapM_ (\x -> a (toRawFilePath x) (toRawFilePath (dstdir </> takeFileName x)))
 			=<< dirContents srcdir
 #endif
 	sanitycheck dir = 
@@ -222,7 +222,7 @@ upgradeToDistribution newdir cleanup distributionfile = do
 	makeorigsymlink olddir = do
 		let origdir = fromRawFilePath (parentDir (toRawFilePath olddir)) </> installBase
 		removeWhenExistsWith R.removeLink (toRawFilePath origdir)
-		createSymbolicLink newdir origdir
+		R.createSymbolicLink (toRawFilePath newdir) (toRawFilePath origdir)
 
 {- Finds where the old version was installed. -}
 oldVersionLocation :: IO FilePath
