@@ -49,9 +49,10 @@ import qualified Data.Set as S
 import qualified Data.Map as M
 import Data.Either
 import qualified System.FilePath.ByteString as P
+import System.PosixCompat.Files (fileMode, isSymbolicLink, modificationTime)
 
 cmd :: Command
-cmd = withGlobalOptions [jobsOption, jsonOptions, annexedMatchingOptions] $
+cmd = withAnnexOptions [jobsOption, jsonOptions, annexedMatchingOptions] $
 	command "fsck" SectionMaintenance
 		"find and fix problems"
 		paramPaths (seek <$$> optParser)
@@ -164,7 +165,7 @@ performRemote key afile backend numcopies remote =
 			Nothing -> go True Nothing
 			Just (Right verification) -> go True (Just (tmpfile, verification))
 			Just (Left _) -> do
-				warning "failed to download file from remote"
+				warning (decodeBS (actionItemDesc ai) ++ ": failed to download file from remote")
 				void $ go True Nothing
 				return False
 	dispatch (Right False) = go False Nothing
@@ -195,7 +196,7 @@ performRemote key afile backend numcopies remote =
 	getfile tmp = ifM (checkDiskSpace (Just (P.takeDirectory tmp)) key 0 True)
 		( ifM (getcheap tmp)
 			( return (Just (Right UnVerified))
-			, ifM (Annex.getState Annex.fast)
+			, ifM (Annex.getRead Annex.fast)
 				( return Nothing
 				, Just <$> tryNonAsync (getfile' tmp)
 				)
@@ -574,7 +575,7 @@ checkKeyNumCopies key afile numcopies = do
 	(deadlocations, safelocations) <- trustPartition DeadTrusted otherlocations
 	let present = length safelocations
 	if present < fromNumCopies numcopies
-		then ifM (pure (not hasafile) <&&> checkDead key)
+		then ifM (checkDead key)
 			( do
 				showLongNote $ "This key is dead, skipping."
 				return True
@@ -666,7 +667,7 @@ recordFsckTime inc key = withFsckDb inc $ \h -> liftIO $ FsckDb.addDb h key
 
 {- Records the start time of an incremental fsck.
  -
- - To guard against time stamp damange (for example, if an annex directory
+ - To guard against time stamp damage (for example, if an annex directory
  - is copied without -a), the fsckstate file contains a time that should
  - be identical to its modification time.
  - (This is not possible to do on Windows, and so the timestamp in
