@@ -16,6 +16,7 @@ import Text.Feed.Query
 import Text.Feed.Types
 import qualified Data.Set as S
 import qualified Data.Map as M
+import Data.Char
 import Data.Time.Clock
 import Data.Time.Format
 import Data.Time.Calendar
@@ -40,6 +41,7 @@ import Command.AddUrl (addUrlFile, downloadRemoteFile, parseDownloadOptions, Dow
 import Annex.UUID
 import Backend.URL (fromUrl)
 import Annex.Content
+import Annex.WorkTree
 import Annex.YoutubeDl
 import Types.MetaData
 import Logs.MetaData
@@ -49,9 +51,10 @@ import Command.AddUrl (addWorkTree, checkRaw)
 import Annex.UntrustedFilePath
 import qualified Annex.Branch
 import Logs
+import qualified Utility.RawFilePath as R
 
 cmd :: Command
-cmd = notBareRepo $
+cmd = notBareRepo $ withAnnexOptions [backendOption] $
 	command "importfeed" SectionCommon "import files from podcast feeds"
 		(paramRepeating paramUrl) (seek <$$> optParser)
 
@@ -91,15 +94,22 @@ getFeed addunlockedmatcher opts cache url = do
 	-- ourselves because it goes out of its way to handle encodings.
 	go tmpf = liftIO (parseFeedFromFile' tmpf) >>= \case
 		Nothing -> debugfeedcontent tmpf "parsing the feed failed"
-		Just f -> case findDownloads url f of
-			[] -> debugfeedcontent tmpf "bad feed content; no enclosures to download"
-			l -> do
-				showEndOk
-				ifM (and <$> mapM (performDownload addunlockedmatcher opts cache) l)
-					( clearFeedProblem url
-					, void $ feedProblem url 
-						"problem downloading some item(s) from feed"
-					)
+		Just f -> do
+			case map sanitizetitle $ decodeBS $ fromFeedText $ getFeedTitle f of
+				"" -> noop
+				t -> showNote ('"' : t ++ "\"")
+			case findDownloads url f of
+				[] -> debugfeedcontent tmpf "bad feed content; no enclosures to download"
+				l -> do
+					showEndOk
+					ifM (and <$> mapM (performDownload addunlockedmatcher opts cache) l)
+						( clearFeedProblem url
+						, void $ feedProblem url 
+							"problem downloading some item(s) from feed"
+						)
+	sanitizetitle c
+		| isControl c = '_'
+		| otherwise = c
 	debugfeedcontent tmpf msg = do
 		feedcontent <- liftIO $ readFile tmpf
 		fastDebug "Command.ImportFeed" $ unlines
@@ -135,7 +145,7 @@ data Cache = Cache
 	}
 
 getCache :: Maybe String -> Annex Cache
-getCache opttemplate = ifM (Annex.getState Annex.force)
+getCache opttemplate = ifM (Annex.getRead Annex.force)
 	( ret S.empty S.empty
 	, do
 		showStart "importfeed" "gathering known urls" (SeekInput [])
@@ -246,12 +256,12 @@ performDownload' started addunlockedmatcher opts cache todownload = case locatio
 		-- to avoid adding it a second time.
 		let quviurl = setDownloader linkurl QuviDownloader
 		checkknown mediaurl $ checkknown quviurl $
-			ifM (Annex.getState Annex.fast <||> pure (relaxedOption (downloadOptions opts)))
+			ifM (Annex.getRead Annex.fast <||> pure (relaxedOption (downloadOptions opts)))
 				( addmediafast linkurl mediaurl mediakey
 				, downloadmedia linkurl mediaurl mediakey
 				)
   where
-	forced = Annex.getState Annex.force
+	forced = Annex.getRead Annex.force
 
 	{- Avoids downloading any items that are already known to be
 	 - associated with a file in the annex, unless forced. -}
@@ -297,7 +307,9 @@ performDownload' started addunlockedmatcher opts cache todownload = case locatio
 	 - to be re-downloaded. -}
 	makeunique url n file = ifM alreadyexists
 		( ifM forced
-			( ifAnnexed (toRawFilePath f) checksameurl tryanother
+			( lookupKey (toRawFilePath f) >>= \case
+				Just k -> checksameurl k
+				Nothing -> tryanother
 			, tryanother
 			)
 		, return $ Just f
@@ -309,7 +321,7 @@ performDownload' started addunlockedmatcher opts cache todownload = case locatio
 				let (d, base) = splitFileName file
 				in d </> show n ++ "_" ++ base
 		tryanother = makeunique url (n + 1) file
-		alreadyexists = liftIO $ isJust <$> catchMaybeIO (getSymbolicLinkStatus f)
+		alreadyexists = liftIO $ isJust <$> catchMaybeIO (R.getSymbolicLinkStatus (toRawFilePath f))
 		checksameurl k = ifM (elem url <$> getUrls k)
 			( return Nothing
 			, tryanother
@@ -507,7 +519,7 @@ feedState url = fromRepo $ gitAnnexFeedState $ fromUrl url Nothing
  -
  - So, it's not safe to use T.unpack to convert that to a String,
  - because later use of that String by eg encodeBS will crash
- - with an encoding error. Use this instad.
+ - with an encoding error. Use this instead.
  -
  - This should not be used on a Text that is read using the
  - filesystem encoding because it does not reverse that encoding.
