@@ -13,6 +13,7 @@ import "mtl" Control.Monad.State.Strict
 import qualified Data.Map.Strict as M
 import qualified Data.Vector as V
 import qualified System.FilePath.ByteString as P
+import System.PosixCompat.Files (isDirectory)
 import Data.Ord
 import qualified Data.Semigroup as Sem
 import Prelude
@@ -28,6 +29,7 @@ import Utility.DiskFree
 import Annex.Content
 import Annex.UUID
 import Annex.CatFile
+import Annex.WorkTree
 import Logs.UUID
 import Logs.Trust
 import Logs.Location
@@ -46,6 +48,7 @@ import qualified Limit
 import Messages.JSON (DualDisp(..), ObjectMap(..))
 import Annex.BloomFilter
 import qualified Command.Unused
+import qualified Utility.RawFilePath as R
 
 -- a named computation that produces a statistic
 type Stat = StatState (Maybe (String, StatState String))
@@ -97,7 +100,7 @@ emptyStatInfo = StatInfo Nothing Nothing M.empty Nothing
 type StatState = StateT StatInfo Annex
 
 cmd :: Command
-cmd = noCommit $ withGlobalOptions [jsonOptions, annexedMatchingOptions] $
+cmd = noCommit $ withAnnexOptions [jsonOptions, annexedMatchingOptions] $
 	command "info" SectionQuery
 		"information about an item or the repository"
 		(paramRepeating paramItem) (seek <$$> optParser)
@@ -162,30 +165,28 @@ autoenableInfo = showCustom "info" (SeekInput []) $ do
 	return True
 
 itemInfo :: InfoOptions -> (SeekInput, String) -> Annex ()
-itemInfo o (si, p) = ifM (isdir p)
+itemInfo o (si, p) = ifM (isdir (toRawFilePath p))
 	( dirInfo o p si
-	, do
-		v <- Remote.byName' p
-		case v of
-			Right r -> remoteInfo o r si
-			Left _ -> do
-				v' <- Remote.nameToUUID' p
-				case v' of
-					Right u -> uuidInfo o u si
-					Left _ -> do
-						relp <- liftIO $ relPathCwdToFile (toRawFilePath p)
-						ifAnnexed relp
-							(fileInfo o (fromRawFilePath relp) si)
-							(treeishInfo o p si)
+	, Remote.byName' p >>= \case
+		Right r -> remoteInfo o r si
+		Left _ -> Remote.nameToUUID' p >>= \case
+			([], _) -> do
+				relp <- liftIO $ relPathCwdToFile (toRawFilePath p)
+				lookupKey relp >>= \case
+					Just k -> fileInfo o (fromRawFilePath relp) si k
+					Nothing -> treeishInfo o p si
+			([u], _) -> uuidInfo o u si
+			(_us, msg) -> noInfo p si msg
 	)
   where
-	isdir = liftIO . catchBoolIO . (isDirectory <$$> getFileStatus)
+	isdir = liftIO . catchBoolIO . (isDirectory <$$> R.getFileStatus)
 
-noInfo :: String -> SeekInput -> Annex ()
-noInfo s si = do
+noInfo :: String -> SeekInput -> String -> Annex ()
+noInfo s si msg = do
 	showStart "info" (encodeBS s) si
-	showNote $ "not a directory or an annexed file or a treeish or a remote or a uuid"
+	showNote msg
 	showEndFail
+	Annex.incError
 
 dirInfo :: InfoOptions -> FilePath -> SeekInput -> Annex ()
 dirInfo o dir si = showCustom (unwords ["info", dir]) si $ do
@@ -202,6 +203,7 @@ treeishInfo o t si = do
 	mi <- getTreeStatInfo o (Git.Ref (encodeBS t))
 	case mi of
 		Nothing -> noInfo t si
+			"not a directory or an annexed file or a treeish or a remote or a uuid"
 		Just i -> showCustom (unwords ["info", t]) si $ do
 			stats <- selStats 
 				(tostats (tree_name:tree_fast_stats False)) 
@@ -238,7 +240,7 @@ uuidInfo o u si = showCustom (unwords ["info", fromUUID u]) si $ do
 
 selStats :: [Stat] -> [Stat] -> Annex [Stat]
 selStats fast_stats slow_stats = do
-	fast <- Annex.getState Annex.fast
+	fast <- Annex.getRead Annex.fast
 	return $ if fast
 		then fast_stats
 		else fast_stats ++ slow_stats
@@ -597,7 +599,7 @@ cachedRepoData = repoData <$> get
 
 getDirStatInfo :: InfoOptions -> FilePath -> Annex StatInfo
 getDirStatInfo o dir = do
-	fast <- Annex.getState Annex.fast
+	fast <- Annex.getRead Annex.fast
 	matcher <- Limit.getMatcher
 	(presentdata, referenceddata, numcopiesstats, repodata) <-
 		Command.Unused.withKeysFilesReferencedIn dir initial
@@ -625,7 +627,7 @@ getDirStatInfo o dir = do
 
 getTreeStatInfo :: InfoOptions -> Git.Ref -> Annex (Maybe StatInfo)
 getTreeStatInfo o r = do
-	fast <- Annex.getState Annex.fast
+	fast <- Annex.getRead Annex.fast
 	-- git lstree filenames start with a leading "./" that prevents
 	-- matching, and also things like --include are supposed to
 	-- match relative to the current directory, which does not make
