@@ -1,7 +1,7 @@
 {- git-annex command
  -
  - Copyright 2011 Joachim Breitner <mail@joachim-breitner.de>
- - Copyright 2011-2021 Joey Hess <id@joeyh.name>
+ - Copyright 2011-2023 Joey Hess <id@joeyh.name>
  -
  - Licensed under the GNU AGPL version 3 or higher.
  -}
@@ -21,7 +21,6 @@ module Command.Sync (
 	commitMsg,
 	pushBranch,
 	updateBranch,
-	syncBranch,
 	updateBranches,
 	seekExportContent,
 	parseUnrelatedHistoriesOption,
@@ -50,6 +49,7 @@ import Config.DynamicConfig
 import Annex.Path
 import Annex.Wanted
 import Annex.Content
+import Annex.WorkTree
 import Command.Get (getKey')
 import qualified Command.Move
 import qualified Command.Export
@@ -59,9 +59,11 @@ import Annex.UUID
 import Logs.UUID
 import Logs.Export
 import Logs.PreferredContent
+import Logs.View
 import Annex.AutoMerge
 import Annex.AdjustedBranch
 import Annex.AdjustedBranch.Merge
+import Annex.View
 import Annex.Ssh
 import Annex.BloomFilter
 import Annex.UpdateInstead
@@ -83,7 +85,7 @@ import qualified Data.ByteString as S
 import Data.Char
 
 cmd :: Command
-cmd = withGlobalOptions [jobsOption] $
+cmd = withAnnexOptions [jobsOption, backendOption] $
 	command "sync" SectionCommon 
 		"synchronize local repository with remotes"
 		(paramRepeating paramRemote) (seek <--< optParser)
@@ -211,7 +213,7 @@ instance DeferredParseClass SyncOptions where
 seek :: SyncOptions -> CommandSeek
 seek o = do
 	prepMerge
-	startConcurrency downloadStages (seek' o)
+	startConcurrency transferStages (seek' o)
 	
 seek' :: SyncOptions -> CommandSeek
 seek' o = do
@@ -309,7 +311,10 @@ merge currbranch mergeconfig o commitmode tomerge = do
 		(b, _) -> autoMergeFrom tomerge b mergeconfig commitmode canresolvemerge
 
 syncBranch :: Git.Branch -> Git.Branch
-syncBranch = Git.Ref.underBase "refs/heads/synced" . fromAdjustedBranch
+syncBranch = Git.Ref.underBase "refs/heads/synced" . origBranch
+
+origBranch :: Git.Branch -> Git.Branch
+origBranch = fromViewBranch . fromAdjustedBranch
 
 remoteBranch :: Remote -> Git.Ref -> Git.Ref
 remoteBranch remote = Git.Ref.underBase $ "refs/remotes/" ++ Remote.name remote
@@ -324,7 +329,7 @@ syncRemotes ps = do
 
 syncRemotes' :: [String] -> [Remote] -> Annex [Remote]
 syncRemotes' ps available = 
-	ifM (Annex.getState Annex.fast) ( fastest <$> wanted , wanted )
+	ifM (Annex.getRead Annex.fast) ( fastest <$> wanted , wanted )
   where
 	wanted
 		| null ps = filterM good (concat $ Remote.byCost available)
@@ -404,10 +409,13 @@ mergeLocal' _ _ currbranch@(Nothing, _) = inRepo Git.Branch.currentUnsafe >>= \c
 
 -- Returns the branch that should be merged, if any.
 needMerge :: CurrBranch -> Git.Branch -> Annex (Maybe Git.Branch)
-needMerge currbranch headbranch = ifM (allM id checks)
-	( return (Just syncbranch)
-	, return Nothing
-	)
+needMerge currbranch headbranch
+	| is_branchView headbranch = return Nothing
+	| otherwise = 
+		ifM (allM id checks)
+			( return (Just syncbranch)
+			, return Nothing
+			)
   where
 	syncbranch = syncBranch headbranch
 	checks = case currbranch of
@@ -431,26 +439,43 @@ pushLocal o b = stopUnless (notOnlyAnnex o) $ do
 updateBranches :: CurrBranch -> Annex ()
 updateBranches (Nothing, _) = noop
 updateBranches (Just branch, madj) = do
-	-- When in an adjusted branch, propigate any changes made to it
-	-- back to the original branch. The adjusted branch may also need
-	-- to be updated, if the adjustment is not stable, and the usual
-	-- configuration does not update it.
-	case madj of
-		Nothing -> noop
-		Just adj -> do
-			let origbranch = branch
-			propigateAdjustedCommits origbranch adj
-			unless (adjustmentIsStable adj) $
-				annexAdjustedBranchRefresh <$> Annex.getGitConfig >>= \case
-					0 -> adjustedBranchRefreshFull adj origbranch
-					_ -> return ()
-					
+	-- When in a view branch, update it to reflect any changes
+	-- of its parent branch or the metadata.
+	currentView >>= \case
+		Just (view, madj') -> updateView view madj' >>= \case
+			Nothing -> noop
+			Just newcommit -> do
+				ok <- inRepo $ Git.Command.runBool
+					[ Param "merge"
+					, Param (Git.fromRef newcommit)
+					]
+				unless ok $
+					giveup $ "failed to update view"
+				case madj' of
+					Nothing -> noop
+					Just adj -> updateadjustedbranch adj
+		-- When in an adjusted branch, propagate any changes
+		-- made to it back to the original branch.
+		Nothing -> case madj of
+			Just adj -> do
+				propigateAdjustedCommits branch adj
+				updateadjustedbranch adj
+			Nothing -> noop
+	
 	-- Update the sync branch to match the new state of the branch
-	inRepo $ updateBranch (syncBranch branch) branch
+	inRepo $ updateBranch (syncBranch branch) (fromViewBranch branch)
+  where
+	-- The adjusted branch may also need to be updated, if the adjustment
+	-- is not stable, and the usual configuration does not update it.
+	updateadjustedbranch adj = unless (adjustmentIsStable adj) $
+		annexAdjustedBranchRefresh <$> Annex.getGitConfig >>= \case
+			0 -> adjustedBranchRefreshFull adj branch
+			_ -> return ()
 
 updateBranch :: Git.Branch -> Git.Branch -> Git.Repo -> IO ()
 updateBranch syncbranch updateto g = 
-	unlessM go $ giveup $ "failed to update " ++ Git.fromRef syncbranch
+	unlessM go $
+		giveup $ "failed to update " ++ Git.fromRef syncbranch
   where
 	go = Git.Command.runBool
 		[ Param "branch"
@@ -496,12 +521,8 @@ importRemote importcontent o remote currbranch
 	| not (pullOption o) || not wantpull = noop
 	| otherwise = case remoteAnnexTrackingBranch (Remote.gitconfig remote) of
 		Nothing -> noop
-		Just tb -> do
-			let (b, p) = separate' (== (fromIntegral (ord ':'))) (Git.fromRef' tb)
-			let branch = Git.Ref b
-			let subdir = if S.null p
-				then Nothing
-				else Just (asTopFilePath p)
+		Just b -> do
+			let (branch, subdir) = splitRemoteAnnexTrackingBranchSubdir b
 			if canImportKeys remote importcontent
 				then do
 					Command.Import.seekRemote remote branch subdir importcontent (CheckGitIgnore True)
@@ -564,7 +585,9 @@ mergeRemote remote currbranch mergeconfig o = ifM isBareRepo
 		(mapM (merge currbranch mergeconfig o Git.Branch.ManualCommit . remoteBranch remote) =<< getlist)
 	tomerge = filterM (changed remote)
 	branchlist Nothing = []
-	branchlist (Just branch) = [fromAdjustedBranch branch, syncBranch branch]
+	branchlist (Just branch)
+		| is_branchView branch = []
+		| otherwise = [origBranch branch, syncBranch branch]
 
 pushRemote :: SyncOptions -> Remote -> CurrBranch -> CommandStart
 pushRemote _o _remote (Nothing, _) = stop
@@ -653,7 +676,7 @@ pushBranch :: Remote -> Maybe Git.Branch -> MessageState -> Git.Repo -> IO Bool
 pushBranch remote mbranch ms g = directpush `after` annexpush `after` syncpush
   where
 	syncpush = flip Git.Command.runBool g $ pushparams $ catMaybes
-		[ (refspec . fromAdjustedBranch) <$> mbranch
+		[ (refspec . origBranch) <$> mbranch
 		, Just $ Git.Branch.forcePush $ refspec Annex.Branch.name
 		]
 	annexpush = void $ tryIO $ flip Git.Command.runQuiet g $ pushparams
@@ -672,7 +695,7 @@ pushBranch remote mbranch ms g = directpush `after` annexpush `after` syncpush
 		-- will want to see that one.
 		Just branch -> do
 			let p = flip Git.Command.gitCreateProcess g $ pushparams
-				[ Git.fromRef $ Git.Ref.base $ fromAdjustedBranch branch ]
+				[ Git.fromRef $ Git.Ref.base $ origBranch branch ]
 			(transcript, ok) <- processTranscript' p Nothing
 			when (not ok && not ("denyCurrentBranch" `isInfixOf` transcript)) $
 				hPutStr stderr transcript
@@ -716,7 +739,7 @@ newer remote b = do
 		)
 
 {- Without --all, only looks at files in the work tree.
- - (Or, when in an ajusted branch where some files are hidden, at files in
+ - (Or, when in an adjusted branch where some files are hidden, at files in
  - the original branch.)
  -
  - With --all, when preferred content expressions look at filenames,
@@ -765,7 +788,10 @@ seekSyncContent o rs currbranch = do
 			seekHelper fst3 ww LsFiles.inRepoDetails l
 
 	seekincludinghidden origbranch mvar l bloomfeeder =
-		seekFiltered (const (pure True)) (\(si, f) -> ifAnnexed f (commandAction . gofile bloomfeeder mvar si f) noop) $
+		let filterer = \(si, f) -> lookupKey f >>= \case
+			Just k -> (commandAction $ gofile bloomfeeder mvar si f k)
+			Nothing -> noop
+		in seekFiltered (const (pure True)) filterer $
 			seekHelper id ww (LsFiles.inRepoOrBranch origbranch) l 
 
 	ww = WarnUnmatchLsFiles
@@ -855,7 +881,7 @@ syncFile ebloom rs af k = do
 	wantput r
 		| Remote.readonly r || remoteAnnexReadOnly (Remote.gitconfig r) = return False
 		| isThirdPartyPopulated r = return False
-		| otherwise = wantSend True (Just k) af (Remote.uuid r)
+		| otherwise = wantGetBy True (Just k) af (Remote.uuid r)
 	handleput lack inhere
 		| inhere = catMaybes <$>
 			( forM lack $ \r ->
@@ -871,9 +897,14 @@ syncFile ebloom rs af k = do
 	ai = mkActionItem (k, af)
 	si = SeekInput []
 
-{- When a remote has an annex-tracking-branch configuration, change the export
- - to contain the current content of the branch. Otherwise, transfer any files
- - that were part of an export but are not in the remote yet.
+{- When a remote has an annex-tracking-branch configuration, and that branch
+ - is currently checked out, change the export to contain the current content
+ - of the branch. (If the branch is not currently checked out, anything
+ - imported from the remote will not yet have been merged into it yet and
+ - so exporting would delete files from the remote unexpectedly.)
+ -
+ - Otherwise, transfer any files that were part of a previous export
+ - but are not in the remote yet.
  - 
  - Returns True if any file transfers were made.
  -}
@@ -888,20 +919,30 @@ seekExportContent o rs (currbranch, _) = or <$> forM rs go
 			Export.closeDb
 			(\db -> Export.writeLockDbWhile db (go' r db))
 	go' r db = case remoteAnnexTrackingBranch (Remote.gitconfig r) of
-		Nothing -> cannotupdateexport r db Nothing
+		Nothing -> cannotupdateexport r db Nothing True
 		Just b -> do
 			mtree <- inRepo $ Git.Ref.tree b
+			let addsubdir = case snd (splitRemoteAnnexTrackingBranchSubdir b) of
+				Just subdir -> \cb -> Git.Ref $
+					Git.fromRef' cb  <> ":" <> getTopFilePath subdir
+				Nothing -> id
+			mcurrtree <- maybe (pure Nothing)
+				(inRepo . Git.Ref.tree . addsubdir)
+				currbranch
 			mtbcommitsha <- Command.Export.getExportCommit r b
-			case (mtree, mtbcommitsha) of
-				(Just tree, Just _) -> do
-					filteredtree <- Command.Export.filterExport r tree
-					Command.Export.changeExport r db filteredtree
-					Command.Export.fillExport r db filteredtree mtbcommitsha
-				_ -> cannotupdateexport r db (Just b)
+			case (mtree, mcurrtree, mtbcommitsha) of
+				(Just tree, Just currtree, Just _)
+					| tree == currtree -> do
+						filteredtree <- Command.Export.filterExport r tree
+						Command.Export.changeExport r db filteredtree
+						Command.Export.fillExport r db filteredtree mtbcommitsha
+					| otherwise -> cannotupdateexport r db (Just b) False
+				_ -> cannotupdateexport r db (Just b) True
 	
-	cannotupdateexport r db mtb = do
+	cannotupdateexport r db mtb showwarning = do
 		exported <- getExport (Remote.uuid r)
-		maybe noop (warncannotupdateexport r mtb exported) currbranch
+		when showwarning $
+			maybe noop (warncannotupdateexport r mtb exported) currbranch
 		fillexistingexport r db (exportedTreeishes exported) Nothing
 	
 	warncannotupdateexport r mtb exported currb = case mtb of
@@ -916,7 +957,7 @@ seekExportContent o rs (currbranch, _) = or <$> forM rs go
 			_ -> noop
 		Just b -> showLongNote $ unwords
 			[ notupdating
-			, "because " ++ Git.fromRef b ++ "does not exist."
+			, "because " ++ Git.fromRef b ++ " does not exist."
 			, "(As configured by " ++ gitconfig ++ ")"
 			]
 	  where
@@ -988,3 +1029,12 @@ isImport = importTree . Remote.config
 
 isThirdPartyPopulated :: Remote -> Bool
 isThirdPartyPopulated = Remote.thirdPartyPopulated . Remote.remotetype
+
+splitRemoteAnnexTrackingBranchSubdir :: Git.Ref -> (Git.Ref, Maybe TopFilePath)
+splitRemoteAnnexTrackingBranchSubdir tb = (branch, subdir)
+  where
+	(b, p) = separate' (== (fromIntegral (ord ':'))) (Git.fromRef' tb)
+	branch = Git.Ref b
+	subdir = if S.null p
+		then Nothing
+		else Just (asTopFilePath p)
