@@ -1,6 +1,6 @@
 {- user-specified limits on files to act on
  -
- - Copyright 2011-2021 Joey Hess <id@joeyh.name>
+ - Copyright 2011-2022 Joey Hess <id@joeyh.name>
  -
  - Licensed under the GNU AGPL version 3 or higher.
  -}
@@ -45,6 +45,7 @@ import Data.Time.Clock.POSIX
 import qualified Data.Set as S
 import qualified Data.Map as M
 import qualified System.FilePath.ByteString as P
+import System.PosixCompat.Files (accessTime, isSymbolicLink)
 
 {- Some limits can look at the current status of files on
  - disk, or in the annex. This allows controlling which happens. -}
@@ -121,7 +122,7 @@ limitExclude glob = Right $ MatchFiles
 matchGlobFile :: String -> MatchInfo -> Annex Bool
 matchGlobFile glob = go
   where
-	cglob = compileGlob glob CaseSensative (GlobFilePath True) -- memoized
+	cglob = compileGlob glob CaseSensitive (GlobFilePath True) -- memoized
 	go (MatchingFile fi) = pure $ matchGlob cglob (fromRawFilePath (matchFile fi))
 	go (MatchingInfo p) = pure $ case providedFilePath p of
 		Just f -> matchGlob cglob (fromRawFilePath f)
@@ -167,7 +168,7 @@ matchSameContentGlob glob mi = checkKey (go mi) mi
 		check k . toRawFilePath
 			=<< getUserInfo (userProvidedFilePath p)
 	
-	cglob = compileGlob glob CaseSensative (GlobFilePath True) -- memoized
+	cglob = compileGlob glob CaseSensitive (GlobFilePath True) -- memoized
 	
 	matchesglob f = matchGlob cglob (fromRawFilePath f)
 #ifdef mingw32_HOST_OS
@@ -231,7 +232,7 @@ matchMagic _limitname querymagic selectprovidedinfo selectuserprovidedinfo (Just
 		, matchNeedsLocationLog = False
 		}
   where
- 	cglob = compileGlob glob CaseSensative (GlobFilePath False) -- memoized
+ 	cglob = compileGlob glob CaseSensitive (GlobFilePath False) -- memoized
 	go (MatchingFile fi) = catchBoolIO $
 		maybe False (matchGlob cglob)
 			<$> querymagic magic (fromRawFilePath (contentFile fi))
@@ -272,7 +273,7 @@ matchLockStatus wantlocked (MatchingFile fi) = liftIO $ do
 	islocked <- isPointerFile f >>= \case
 		Just _key -> return False
 		Nothing -> isSymbolicLink
-			<$> getSymbolicLinkStatus (fromRawFilePath f)
+			<$> R.getSymbolicLinkStatus f
 	return (islocked == wantlocked)
 matchLockStatus wantlocked (MatchingInfo p) = 
 	pure $ case providedLinkType p of
@@ -282,7 +283,7 @@ matchLockStatus wantlocked (MatchingInfo p) =
 matchLockStatus _ (MatchingUserInfo _) = pure False
 
 {- Adds a limit to skip files not believed to be present
- - in a specfied repository. Optionally on a prior date. -}
+ - in a specified repository. Optionally on a prior date. -}
 addIn :: String -> Annex ()
 addIn s = do
 	u <- Remote.nameToUUID name
@@ -431,6 +432,10 @@ limitUnused = MatchFiles
 	
 	isunused k = S.member k <$> unusedKeys
 
+{- Adds a limit that matches anything. -}
+addAnything :: Annex ()
+addAnything = addLimit (Right limitAnything)
+
 {- Limit that matches any version of any file or key. -}
 limitAnything :: MatchFiles Annex
 limitAnything = MatchFiles
@@ -441,6 +446,10 @@ limitAnything = MatchFiles
 	, matchNeedsLocationLog = False
 	}
 
+{- Adds a limit that never matches. -}
+addNothing :: Annex ()
+addNothing = addLimit (Right limitNothing)
+
 {- Limit that never matches. -}
 limitNothing :: MatchFiles Annex
 limitNothing = MatchFiles 
@@ -572,7 +581,7 @@ addAccessedWithin duration = do
   where
 	check now k = inAnnexCheck k $ \f ->
 		liftIO $ catchDefaultIO False $ do
-			s <- R.getFileStatus f
+			s <- R.getSymbolicLinkStatus f
 			let accessed = realToFrac (accessTime s)
 			let delta = now - accessed
 			return $ delta <= secs
