@@ -19,6 +19,7 @@ import qualified Data.ByteString.Lazy as L
 import qualified Data.Map as M
 import qualified System.FilePath.ByteString as P
 import Data.Default
+import System.PosixCompat.Files (isRegularFile, getFdStatus, deviceID)
 
 import Annex.Common
 import Types.Remote
@@ -73,7 +74,7 @@ ignoreinodesField = Accepted "ignoreinodes"
 gen :: Git.Repo -> UUID -> RemoteConfig -> RemoteGitConfig -> RemoteStateHandle -> Annex (Maybe Remote)
 gen r u rc gc rs = do
 	c <- parsedRemoteConfig remote rc
-	cst <- remoteCost gc cheapRemoteCost
+	cst <- remoteCost gc c cheapRemoteCost
 	let chunkconfig = getChunkConfig c
 	cow <- liftIO newCopyCoWTried
 	let ii = IgnoreInodes $ fromMaybe True $
@@ -152,7 +153,7 @@ directorySetup _ mu _ c gc = do
 	(c', _encsetup) <- encryptionSetup c gc
 
 	-- The directory is stored in git config, not in this remote's
-	-- persistant state, so it can vary between hosts.
+	-- persistent state, so it can vary between hosts.
 	gitConfigSpecialRemote u c' [("directory", absdir)]
 	return (M.delete directoryField c', u)
 
@@ -182,7 +183,7 @@ storeKeyM :: RawFilePath -> ChunkConfig -> CopyCoWTried -> Storer
 storeKeyM d chunkconfig cow k c m = 
 	ifM (checkDiskSpaceDirectory d k)
 		( do
-			void $ liftIO $ tryIO $ createDirectoryUnder d tmpdir
+			void $ liftIO $ tryIO $ createDirectoryUnder [d] tmpdir
 			store
 		, giveup "Not enough free disk space."
 		)
@@ -218,8 +219,8 @@ checkDiskSpaceDirectory d k = do
 	annexdir <- fromRepo gitAnnexObjectDir
 	samefilesystem <- liftIO $ catchDefaultIO False $ 
 		(\a b -> deviceID a == deviceID b)
-			<$> R.getFileStatus d
-			<*> R.getFileStatus annexdir
+			<$> R.getSymbolicLinkStatus d
+			<*> R.getSymbolicLinkStatus annexdir
 	checkDiskSpace (Just d) k 0 samefilesystem
 
 {- Passed a temp directory that contains the files that should be placed
@@ -229,7 +230,7 @@ checkDiskSpaceDirectory d k = do
 finalizeStoreGeneric :: RawFilePath -> RawFilePath -> RawFilePath -> IO ()
 finalizeStoreGeneric d tmp dest = do
 	removeDirGeneric (fromRawFilePath d) dest'
-	createDirectoryUnder d (parentDir dest)
+	createDirectoryUnder [d] (parentDir dest)
 	renameDirectory (fromRawFilePath tmp) dest'
 	-- may fail on some filesystems
 	void $ tryIO $ do
@@ -254,7 +255,7 @@ retrieveKeyFileCheapM _ (LegacyChunks _) = Nothing
 retrieveKeyFileCheapM d NoChunks = Just $ \k _af f -> liftIO $ do
 	file <- fromRawFilePath <$> (absPath =<< getLocation d k)
 	ifM (doesFileExist file)
-		( createSymbolicLink file f
+		( R.createSymbolicLink (toRawFilePath file) (toRawFilePath f)
 		, giveup "content file not present in remote"
 		)
 #else
@@ -309,7 +310,7 @@ checkPresentGeneric' d check = ifM check
 
 storeExportM :: RawFilePath -> CopyCoWTried -> FilePath -> Key -> ExportLocation -> MeterUpdate -> Annex ()
 storeExportM d cow src _k loc p = do
-	liftIO $ createDirectoryUnder d (P.takeDirectory dest)
+	liftIO $ createDirectoryUnder [d] (P.takeDirectory dest)
 	-- Write via temp file so that checkPresentGeneric will not
 	-- see it until it's fully stored.
 	viaTmp go (fromRawFilePath dest) ()
@@ -337,7 +338,7 @@ checkPresentExportM d _k loc =
 
 renameExportM :: RawFilePath -> Key -> ExportLocation -> ExportLocation -> Annex (Maybe ())
 renameExportM d _k oldloc newloc = liftIO $ do
-	createDirectoryUnder d (P.takeDirectory dest)
+	createDirectoryUnder [d] (P.takeDirectory dest)
 	renameFile (fromRawFilePath src) (fromRawFilePath dest)
 	removeExportLocation d oldloc
 	return (Just ())
@@ -363,13 +364,13 @@ removeExportLocation topdir loc =
 
 listImportableContentsM :: IgnoreInodes -> RawFilePath -> Annex (Maybe (ImportableContentsChunkable Annex (ContentIdentifier, ByteSize)))
 listImportableContentsM ii dir = liftIO $ do
-	l <- dirContentsRecursive (fromRawFilePath dir)
+	l <- dirContentsRecursiveSkipping (const False) False (fromRawFilePath dir)
 	l' <- mapM (go . toRawFilePath) l
 	return $ Just $ ImportableContentsComplete $
 		ImportableContents (catMaybes l') []
   where
 	go f = do
-		st <- R.getFileStatus f
+		st <- R.getSymbolicLinkStatus f
 		mkContentIdentifier ii f st >>= \case
 			Nothing -> return Nothing
 			Just cid -> do
@@ -390,10 +391,24 @@ mkContentIdentifier (IgnoreInodes ii) f st =
 			then toInodeCache' noTSDelta f st 0
 			else toInodeCache noTSDelta f st
 
-guardSameContentIdentifiers :: a -> ContentIdentifier -> Maybe ContentIdentifier -> a
-guardSameContentIdentifiers cont old new
-	| new == Just old = cont
+-- Since ignoreinodes can be changed by enableremote, and since previous
+-- versions of git-annex ignored inodes by default, treat two content
+-- idenfiers as the same if they differ only by one having the inode
+-- ignored.
+guardSameContentIdentifiers :: a -> [ContentIdentifier] -> Maybe ContentIdentifier -> a
+guardSameContentIdentifiers _ _ Nothing = giveup "file not found"
+guardSameContentIdentifiers cont olds (Just new)
+	| any (new ==) olds = cont
+	| any (ignoreinode new ==) olds = cont
+	| any (\old -> new == ignoreinode old) olds = cont
 	| otherwise = giveup "file content has changed"
+  where
+	ignoreinode cid@(ContentIdentifier b) = 
+		case readInodeCache (decodeBS b) of
+			Nothing -> cid
+			Just ic -> 
+				let ic' = replaceInode 0 ic
+				in ContentIdentifier (encodeBS (showInodeCache ic'))
 
 importKeyM :: IgnoreInodes -> RawFilePath -> ExportLocation -> ContentIdentifier -> ByteSize -> MeterUpdate -> Annex (Maybe Key)
 importKeyM ii dir loc cid sz p = do
@@ -402,8 +417,8 @@ importKeyM ii dir loc cid sz p = do
 	let k = alterKey unsizedk $ \kd -> kd
 		{ keySize = keySize kd <|> Just sz }
 	currcid <- liftIO $ mkContentIdentifier ii absf
-		=<< R.getFileStatus absf
-	guardSameContentIdentifiers (return (Just k)) cid currcid
+		=<< R.getSymbolicLinkStatus absf
+	guardSameContentIdentifiers (return (Just k)) [cid] currcid
   where
 	f = fromExportLocation loc
 	absf = dir P.</> f
@@ -413,8 +428,8 @@ importKeyM ii dir loc cid sz p = do
 		, inodeCache = Nothing
 		}
 
-retrieveExportWithContentIdentifierM :: IgnoreInodes -> RawFilePath -> CopyCoWTried -> ExportLocation -> ContentIdentifier -> FilePath -> Either Key (Annex Key) -> MeterUpdate -> Annex (Key, Verification)
-retrieveExportWithContentIdentifierM ii dir cow loc cid dest gk p =
+retrieveExportWithContentIdentifierM :: IgnoreInodes -> RawFilePath -> CopyCoWTried -> ExportLocation -> [ContentIdentifier] -> FilePath -> Either Key (Annex Key) -> MeterUpdate -> Annex (Key, Verification)
+retrieveExportWithContentIdentifierM ii dir cow loc cids dest gk p =
 	case gk of
 		Right mkkey -> do
 			go Nothing
@@ -460,9 +475,9 @@ retrieveExportWithContentIdentifierM ii dir cow loc cid dest gk p =
 	
 	-- Check before copy, to avoid expensive copy of wrong file
 	-- content.
-	precheck cont = guardSameContentIdentifiers cont cid
+	precheck cont = guardSameContentIdentifiers cont cids
 		=<< liftIO . mkContentIdentifier ii f
-		=<< liftIO (R.getFileStatus f)
+		=<< liftIO (R.getSymbolicLinkStatus f)
 
 	-- Check after copy, in case the file was changed while it was
 	-- being copied.
@@ -486,9 +501,9 @@ retrieveExportWithContentIdentifierM ii dir cow loc cid dest gk p =
 #ifndef mingw32_HOST_OS
 			=<< getFdStatus fd
 #else
-			=<< R.getFileStatus f
+			=<< R.getSymbolicLinkStatus f
 #endif
-		guardSameContentIdentifiers cont cid currcid
+		guardSameContentIdentifiers cont cids currcid
 
 	-- When copy-on-write was done, cannot check the handle that was
 	-- copied from, but such a copy should run very fast, so
@@ -497,28 +512,28 @@ retrieveExportWithContentIdentifierM ii dir cow loc cid dest gk p =
 	-- restored to the original content before this check.
 	postcheckcow cont = do
 		currcid <- liftIO $ mkContentIdentifier ii f
-			=<< R.getFileStatus f
-		guardSameContentIdentifiers cont cid currcid
+			=<< R.getSymbolicLinkStatus f
+		guardSameContentIdentifiers cont cids currcid
 
 storeExportWithContentIdentifierM :: IgnoreInodes -> RawFilePath -> CopyCoWTried -> FilePath -> Key -> ExportLocation -> [ContentIdentifier] -> MeterUpdate -> Annex ContentIdentifier
 storeExportWithContentIdentifierM ii dir cow src _k loc overwritablecids p = do
-	liftIO $ createDirectoryUnder dir (toRawFilePath destdir)
+	liftIO $ createDirectoryUnder [dir] (toRawFilePath destdir)
 	withTmpFileIn destdir template $ \tmpf tmph -> do
 		liftIO $ hClose tmph
 		void $ liftIO $ fileCopier cow src tmpf p Nothing
 		let tmpf' = toRawFilePath tmpf
 		resetAnnexFilePerm tmpf'
-		liftIO (getFileStatus tmpf) >>= liftIO . mkContentIdentifier ii tmpf' >>= \case
+		liftIO (R.getSymbolicLinkStatus tmpf') >>= liftIO . mkContentIdentifier ii tmpf' >>= \case
 			Nothing -> giveup "unable to generate content identifier"
 			Just newcid -> do
 				checkExportContent ii dir loc
 					overwritablecids
 					(giveup "unsafe to overwrite file")
-					(const $ liftIO $ rename tmpf dest)
+					(const $ liftIO $ R.rename tmpf' dest)
 				return newcid
   where
-	dest = fromRawFilePath $ exportPath dir loc
-	(destdir, base) = splitFileName dest
+	dest = exportPath dir loc
+	(destdir, base) = splitFileName (fromRawFilePath dest)
 	template = relatedTemplate (base ++ ".tmp")
 
 removeExportWithContentIdentifierM :: IgnoreInodes -> RawFilePath -> Key -> ExportLocation -> [ContentIdentifier] -> Annex ()
@@ -553,7 +568,7 @@ data CheckResult = DoesNotExist | KnownContentIdentifier
 -- content is known, and immediately run the callback.
 checkExportContent :: IgnoreInodes -> RawFilePath -> ExportLocation -> [ContentIdentifier] -> Annex a -> (CheckResult -> Annex a) -> Annex a
 checkExportContent ii dir loc knowncids unsafe callback = 
-	tryWhenExists (liftIO $ R.getFileStatus dest) >>= \case
+	tryWhenExists (liftIO $ R.getSymbolicLinkStatus dest) >>= \case
 		Just destst
 			| not (isRegularFile destst) -> unsafe
 			| otherwise -> catchDefaultIO Nothing (liftIO $ mkContentIdentifier ii dest destst) >>= \case
